#!/usr/bin/env python

# Twister sprite

import os
import pygame
from pygame.locals import *

from vector import Vector2D
from mediator import mediator

FRICTION = 0.016
TWIST_LOSS = 0.018
TERRAIN_COST = 0.01

class Twister(pygame.sprite.Sprite):
    def __init__(self, initPos, radius, moveVector=(0,0), twist=0,
                 terrain=None, mediator=mediator):
        """
        Create a new Twister sprite.
        
        `initPos`    2-tuple defining the origin of the twister
        `radius`     radius of the twister
        `moveVector` 2-tuple vector defining twister speed (pixels/frame)
        `twist`      rotatation amount
        """
        super(Twister, self).__init__()
        self.moveVector = Vector2D(*moveVector)
        self.dia = radius * 2
        img = pygame.image.load(os.path.join("data", "tornado2.png"))
        self.timg = pygame.transform.scale(
            img, (self.dia, self.dia))
        self.timg2 = pygame.transform.scale(
            img, (self.dia/2, self.dia/2))
        self.timg3 = pygame.transform.scale(
            img, (self.dia/3, self.dia/3))
        self.rect = pygame.rect.Rect(0, 0, self.dia, self.dia)
        self.position = Vector2D(*initPos)
        self.twist = float(twist)
        self.curTwist = 0
        self.terrain = terrain
        self.mediator = mediator
        self.nFrames = 0
        if (self.mediator):
            self.mediator.addListener('tick', self.tickHandler)
            self.mediator.publishEvent('twister_spawned', twister=self)

    def position():
        doc = "XY position of twister's origin (using Vector2D)"
        def fget(self):
            return Vector2D(*self.rect.center)
        def fset(self, pos):
            self.rect.center = (int(pos.x), int(pos.y))
        return locals()
    position = property(**position())
    
    def radius():
        doc = "twister radius"
        def fget(self):
            return self.dia / 2.0
        return locals()
    radius = property(**radius())
    
    def image():
        doc = "Sprite image"
        def fget(self):
            r = self.dia / 2
            surf = pygame.transform.rotate(self.timg, self.curTwist)
            surf.blit(pygame.transform.rotate(self.timg, -self.curTwist / 2),
                (0,0))
            surf.blit(pygame.transform.rotate(self.timg2, -self.curTwist / 3),
                (self.dia * 0.25,self.dia * 0.25))
            surf.blit(pygame.transform.rotate(self.timg3, self.curTwist / 4),
                (self.dia * 0.33, self.dia * 0.33))
            return surf
        return locals()
    image = property(**image())
    
    def moveTwister(self):
        """This should be called every frame to update the sprite state."""
        self.nFrames += 1
        oldPos = self.position
        self.position += self.moveVector
        self.moveVector = self.moveVector.rotate(
            (self.twist / 1000.0) * (self.nFrames / 320.0))
        self.twist *= 1.0 - (TWIST_LOSS * self.nFrames / 240.0)
        self.moveVector *= 1.0 - FRICTION
        
        self.curTwist += self.twist
        self.curTwist %= 1440
        
        if (self.terrain):
            t = self.terrain.t
            p0 = oldPos
            p1 = self.position
            try:
                self.moveVector += Vector2D(
                    TERRAIN_COST * (t[p0.x][p1.y] - t[p1.x][p1.y]),
                    TERRAIN_COST * (t[p1.y][p0.y] - t[p1.x][p1.y]))
            except IndexError:
                # Gone off the map
                pass
        
        if (self.mediator):
            self.mediator.publishEvent("twister_moved", twister=self)
        
        if ((self.moveVector.length < 0.25) and self.mediator):
            self.mediator.removeListener('tick', self.tickHandler)
            self.mediator.publishEvent('twister_died', twister=self)
    
    def tickHandler(self, event):
        self.moveTwister()

    def seppuku(self):
        print "Twister performing seppuku"
        self.mediator.removeListener('tick', self.tickHandler)
        

# Some test code to animate a twister
if __name__=='__main__':
    t = Twister((20, 20), 32, (12,0), twist=30)
    t2 = Twister((70, 50), 64, (14,5), twist=30)
    t3 = Twister((200, 80), 128, (5,11), twist=20)
    bg = pygame.image.load(os.path.join("data","bgtest.jpg"))
    screen = pygame.display.set_mode([800, 600])
    while pygame.event.poll().type != KEYDOWN:
        pygame.time.delay(10)
        mediator.publishEvent('tick')
        screen.blit(bg, (0,0))
        for s in (t, t2, t3):
            screen.blit(s.image, s.rect)
        pygame.display.update()
